import argparse
import gzip
import io
import os
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path

try:
    import brotli
except ModuleNotFoundError:  # no cov
    brotli = None

# Extensions that it's not worth trying to compress
SKIP_COMPRESS_EXTENSIONS = (
    # Images
    ".jpg",
    ".jpeg",
    ".png",
    ".gif",
    ".webp",
    # Compressed files
    ".zip",
    ".gz",
    ".tgz",
    ".bz2",
    ".tbz",
    ".xz",
    ".br",
    # Flash
    ".swf",
    ".flv",
    # Fonts
    ".woff",
    ".woff2",
    # Video
    ".3gp",
    ".3gpp",
    ".asf",
    ".avi",
    ".m4v",
    ".mov",
    ".mp4",
    ".mpeg",
    ".mpg",
    ".webm",
    ".wmv",
)


def _write_if_smaller(path, orig_bytes, compress_bytes, algorithm, suffix):
    orig_len = len(orig_bytes)
    compress_len = len(compress_bytes)
    if compress_len < orig_len * 0.9:
        compress_improvement = (compress_len - orig_len) / orig_len
        Path(str(path) + suffix).write_bytes(compress_bytes)
        return f"{path!s}: {algorithm} compressed {orig_len} to {compress_len} bytes ({int(100 * compress_improvement)}%)"
    return f"{path!s}: {algorithm} didn't produce useful compression results"


def try_gzip(path, orig_bytes):
    with io.BytesIO() as f:
        with gzip.GzipFile(
            filename="", mode="wb", fileobj=f, compresslevel=9, mtime=0
        ) as compress_file:
            compress_file.write(orig_bytes)
        return _write_if_smaller(
            path,
            orig_bytes,
            f.getvalue(),
            "Gzip",
            ".gz",
        )


def try_brotli(path, orig_bytes):
    if not brotli:  # no cov
        return ""
    return _write_if_smaller(
        path,
        orig_bytes,
        brotli.compress(orig_bytes),
        "Brotli",
        ".br",
    )


def _compress_path(path):
    orig_bytes = path.read_bytes()
    return (
        try_brotli(path, orig_bytes),
        try_gzip(path, orig_bytes),
    )


def _paths(root):
    for dir_, _dirs, files in os.walk(root):
        dir = Path(dir_)
        for filename in files:
            path = dir / filename
            if path.suffix not in SKIP_COMPRESS_EXTENSIONS:
                yield path


def compress(root):
    with ThreadPoolExecutor() as executor:
        for result in executor.map(_compress_path, _paths(root)):
            print("\n".join(filter(None, result)))
    return 0


def parse_args(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument("root", help="Path containing static files to compress")
    return parser.parse_args(args)


if __name__ == "__main__":
    args = parse_args()
    raise SystemExit(compress(args.root))



import os

from starlette.datastructures import Headers
from starlette.responses import FileResponse, PlainTextResponse

# Ten years is what nginx sets as max age if you use 'expires max;'
# so we'll follow its lead
FOREVER = f"max-age={10 * 365 * 24 * 60 * 60}, public, immutable"
A_LITTE_WHILE = "max-age=60, public"
SUFFIX_ENCODINGS = {".br": "br", ".gz": "gzip"}


def never(_path):
    return False  # no cov


class BlackNoise:
    def __init__(self, application, *, immutable_file_test=never):
        self._files = {}
        self._prefixes = ()
        self._application = application

        self._immutable_file_test = immutable_file_test

    def add(self, path, prefix):
        self._prefixes = (*self._prefixes, prefix)

        for base, _dirs, files in os.walk(path):
            path_prefix = os.path.join(prefix, base[len(str(path)) :].strip("/"))
            self._files |= {
                os.path.join(path_prefix, file): os.path.join(base, file)
                for file in files
                if all(
                    # File is not a compressed file
                    not file.endswith(suffix)
                    # The uncompressed variant does not exist
                    or file.removesuffix(suffix) not in files
                    for suffix in SUFFIX_ENCODINGS
                )
            }

    async def __call__(self, scope, receive, send):
        path = os.path.normpath(scope["path"].removeprefix(scope["root_path"]))

        if scope["type"] != "http" or not path.startswith(self._prefixes):
            response = self._application

        elif scope["method"] not in ("GET", "HEAD"):
            response = PlainTextResponse("Method Not Allowed", status_code=405)

        elif file := self._files.get(path):
            response = await _file_response(
                scope, file, self._immutable_file_test(path)
            )

        else:
            response = PlainTextResponse("Not Found", status_code=404)

        await response(scope, receive, send)


async def _file_response(scope, file, immutable):
    headers = {
        "accept-ranges": "bytes",
        "access-control-allow-origin": "*",
        "cache-control": FOREVER if immutable else A_LITTE_WHILE,
    }
    h = Headers(scope=scope)

    # Defer to Starlette when we get a HTTP Range request.
    if h.get("range"):
        return FileResponse(file, headers=headers)

    accept_encoding = h.get("accept-encoding", "")
    for suffix, encoding in SUFFIX_ENCODINGS.items():
        if encoding not in accept_encoding:
            continue

        if file.endswith(suffix):
            return FileResponse(file, headers=headers | {"content-encoding": encoding})

        compressed_file = f"{file}{suffix}"
        if os.path.exists(compressed_file):
            return FileResponse(
                compressed_file,
                headers=headers | {"content-encoding": encoding},
            )

    return FileResponse(file, headers=headers)



import tempfile
from pathlib import Path

import httpx
import pytest
from httpx_ws import aconnect_ws
from httpx_ws.transport import ASGIWebSocketTransport
from starlette.applications import Starlette
from starlette.responses import PlainTextResponse
from starlette.routing import Route, WebSocketRoute

from blacknoise import BlackNoise
from blacknoise.compress import compress, parse_args


async def http_hello(request):
    return PlainTextResponse(f"Hello from {request.url.path}")


async def ws_hello(websocket):
    await websocket.accept()
    await websocket.send_text("Hello World!")
    await websocket.close()


app = Starlette(
    routes=[
        Route("/http", http_hello),
        WebSocketRoute("/ws", ws_hello),
    ]
)


@pytest.fixture
def bn():
    this = Path(__file__).parent
    blacknoise = BlackNoise(app, immutable_file_test=lambda path: "hello3" in path)
    blacknoise.add(this / "static", "/static/")
    return blacknoise


def test_files_contents(bn):
    this = Path(__file__).parent

    assert "/static/hello.txt" in bn._files
    assert "/static/hello.txt.gz" not in bn._files
    assert "/static/hello2.txt.gz" in bn._files
    assert "/static/hello3.txt" in bn._files
    assert "/static/foo" not in bn._files

    assert bn._files["/static/hello.txt"] == str(this / "static" / "hello.txt")


@pytest.mark.asyncio
async def test_static_file_serving(bn):
    transport = httpx.ASGITransport(app=bn)
    async with httpx.AsyncClient(
        transport=transport, base_url="http://testserver"
    ) as client:
        r = await client.get("/static/hello.txt")
        assert r.status_code == 200
        assert r.text == "world\n"
        assert r.headers["content-encoding"] == "gzip"
        assert r.headers["cache-control"] == "max-age=60, public"
        assert r.headers["access-control-allow-origin"] == "*"

        r = await client.get("/static/hello2.txt.gz")
        assert r.status_code == 200
        assert r.text == "world2\n"
        assert r.headers["content-encoding"] == "gzip"
        assert r.headers["cache-control"] == "max-age=60, public"
        assert r.headers["access-control-allow-origin"] == "*"

        r = await client.get("/static/hello3.txt")
        assert r.status_code == 200
        assert r.text == "world3\n"
        assert "content-encoding" not in r.headers
        assert r.headers["cache-control"] == "max-age=315360000, public, immutable"
        assert r.headers["access-control-allow-origin"] == "*"

        r = await client.post("/static/hello.txt")
        assert r.status_code == 405

        r = await client.get("/static/foo")
        assert r.status_code == 404

        r = await client.get("/http")
        assert r.status_code == 200
        assert r.text == "Hello from /http"


@pytest.mark.asyncio
async def test_accept_encoding(bn):
    transport = httpx.ASGITransport(app=bn)
    async with httpx.AsyncClient(
        transport=transport, base_url="http://testserver"
    ) as client:
        r = await client.get("/static/hello.txt", headers={"accept-encoding": "gzip"})
        assert r.status_code == 200
        assert r.text == "world\n"
        assert r.headers["content-encoding"] == "gzip"

        r = await client.get(
            "/static/hello.txt", headers={"accept-encoding": "identity"}
        )
        assert r.status_code == 200
        assert r.text == "world\n"
        assert "content-encoding" not in r.headers


@pytest.mark.asyncio
async def test_ws(bn):
    async with httpx.AsyncClient(transport=ASGIWebSocketTransport(bn)) as client:
        http_response = await client.get("http://server/http")
        assert http_response.status_code == 200

        async with aconnect_ws("http://server/ws", client) as ws:
            message = await ws.receive_text()
            assert message == "Hello World!"


def test_compress():
    with tempfile.TemporaryDirectory() as root_:
        root = Path(root_)

        (root / "hello.txt").write_text("hello " * 100)
        (root / "hello2.txt").write_text("hello")
        (root / "hello3.jpeg").write_text("")
        compress(root)

        assert {file.name for file in root.glob("*")} == {
            "hello.txt",
            "hello.txt.gz",
            "hello.txt.br",
            "hello2.txt",
            "hello3.jpeg",
        }


def test_parse_args():
    with pytest.raises(SystemExit):
        parse_args([])

    args = parse_args(["hello"])
    assert args.root == "hello"


@pytest.mark.asyncio
async def test_range(bn):
    transport = httpx.ASGITransport(app=bn)
    async with httpx.AsyncClient(
        transport=transport, base_url="http://testserver"
    ) as client:
        r = await client.get("/static/hello.txt", headers={"range": "words=1-2"})
        assert r.status_code == 400
        # assert r.text == "world\n"

        r = await client.get("/static/hello.txt", headers={"range": "bytes=2-1"})
        assert r.status_code == 206
        assert r.text == ""

        r = await client.get("/static/hello.txt", headers={"range": "bytes=1-2"})
        assert r.status_code == 206
        assert r.text == "or"
        assert r.headers["content-range"] == "bytes 1-2/6"

        r = await client.get("/static/hello.txt", headers={"range": "bytes=-2"})
        assert r.status_code == 206
        assert r.text == "d\n"
        assert r.headers["content-range"] == "bytes 4-5/6"
